(* Reading files, scanning, and parsing.

The parser was written naively, using the methodology for SML parser
construction taught to first-year students at DTU, hence its length.
*)

structure Parse :> Parse = struct

local
  open Data;

  fun mkiib opr = IIB (fn (n1, n2) => if opr (n1,n2) then 1 else 0)

  datatype terminal = 
      NAME of string 
    | INT of int
    | LET | LETREC | IN | CASE | OF | IF | THEN | ELSE | LAMBDA | PACK | END
    | EQUAL | ADD | SUB | MUL | DIV | MOD | LE | GE | NE | LT | GT | AMPERSAND
    | LPAR | RPAR | LBRACE | RBRACE | SEMI | COMMA | BAR | DOT | DASHARROW 

  exception Parseerror

  fun check (t, ts) =
      case ts of 
          t1 :: tr => if t1 = t then tr else raise Parseerror
        | _        => raise Parseerror

  fun expr ts = 
      case ts of 
          LET :: tr    => 
              let val (ts1, ds) = defns tr
                  val ts2 = check(IN, ts1)
                  val (ts3, e) = expr ts2
              in (ts3, LetS(ds, e)) end
        | LETREC :: tr => 
              let val (ts1, ds) = defns tr
                  val ts2 = check(IN, ts1)
                  val (ts3, e) = expr ts2
              in (ts3, LetrecS(ds, e)) end
        | CASE :: tr   => 
              let val (ts1, e0) = expr tr 
                  val ts2 = check(OF, ts1)
                  val (ts3, alts1) = alts ts2
                  val ts4 = check(END, ts3)
              in (ts4, CaseS(e0, alts1)) end
        | IF :: tr     => 
              let val (ts1, e1) = expr tr
                  val ts2 = check(THEN, ts1)
                  val (ts3, e2) = expr ts2
                  val ts4 = check(ELSE, ts3)
                  val (ts5, e3) = expr ts4
              in (ts5, IfS(e1, e2, e3)) end
        | LAMBDA :: tr => 
              let val (ts1, n) = (case tr of
                                      NAME n :: tr => (tr, n)
                                    | _            => raise Parseerror)
                  val ts2 = check(DOT, ts1)
                  val (ts3, e) = expr ts2
              in (ts3, LamS(n, e)) end
        | _ => expr1 ts
  and expr1 ts = e1opt (expr2 ts)
  and e1opt (ts, e1) = 
      case ts of
          BAR :: tr => 
              let val (ts1, e2) = expr2 tr
              in e1opt(ts1, IfS(e1, CstS (Bool true), e2)) end
        | _ => (ts, e1)
  and expr2 ts = e2opt (expr3 ts)
  and e2opt (ts, e1) = 
      case ts of
          AMPERSAND :: tr => 
              let val (ts1, e2) = expr3 tr
              in e2opt(ts1, IfS(e1, e2, CstS (Bool false))) end
        | _ => (ts, e1)
  and expr3 ts = e3opt (expr4 ts)
  and e3opt (ts, e1) = 
      case ts of
          EQUAL :: tr => 
              let val (ts1, e2) = expr4 tr
              in (ts1, Op2S(mkiib (op = ), e1, e2)) end
        | NE :: tr => 
              let val (ts1, e2) = expr4 tr
              in (ts1, Op2S(mkiib (op <> ), e1, e2)) end
        | LE :: tr => 
              let val (ts1, e2) = expr4 tr
              in (ts1, Op2S(mkiib (op <= ), e1, e2)) end
        | GE :: tr => 
              let val (ts1, e2) = expr4 tr
              in (ts1, Op2S(mkiib (op >= ), e1, e2)) end
        | LT :: tr => 
              let val (ts1, e2) = expr4 tr
              in (ts1, Op2S(mkiib (op < ), e1, e2)) end
        | GT :: tr => 
              let val (ts1, e2) = expr4 tr
              in (ts1, Op2S(mkiib (op > ), e1, e2)) end
        | _ => (ts, e1)
  and expr4 ts = e4opt (expr5 ts)
  and e4opt (ts, e1) = 
      case ts of
          ADD :: tr => 
              let val (ts1, e2) = expr5 tr
              in e4opt(ts1, Op2S(III (op + ), e1, e2)) end
        | SUB :: tr => 
              let val (ts1, e2) = expr5 tr
              in e4opt(ts1, Op2S(III (op - ), e1, e2)) end
        | _ => (ts, e1)
  and expr5 ts = e5opt (expr6 ts)
  and e5opt (ts, e1) = 
      case ts of
          MUL :: tr => 
              let val (ts1, e2) = expr6 tr
              in e5opt(ts1, Op2S(III (op * ), e1, e2)) end
        | DIV :: tr => 
              let val (ts1, e2) = expr6 tr
              in e5opt(ts1, Op2S(III (op div ), e1, e2)) end
        | MOD :: tr => 
              let val (ts1, e2) = expr6 tr
              in e5opt(ts1, Op2S(III (op mod ), e1, e2)) end
        | _ => (ts, e1)
  and expr6 ts = e6opt (aexpr ts)
  and e6opt (ts, e1) =
      let fun aexpr_e6opt ts = 
              let val (ts1, e2) = aexpr ts
              in e6opt(ts1, AppS(e1, e2)) end
      in case ts of
          NAME _ :: _ => aexpr_e6opt ts
        | INT  _ :: _ => aexpr_e6opt ts
        | PACK   :: _ => aexpr_e6opt ts
        | LPAR   :: _ => aexpr_e6opt ts
        | _           => (ts, e1)
      end
  and aexpr ts = 
      case ts of
          NAME n :: tr => (tr, VarS n)
        | INT  i :: tr => (tr, CstS (Int i))
        | PACK :: LBRACE :: INT k :: tr => 
	      let val (ts1, es) = exprs tr
		  val ts2 = check(RBRACE, ts1)
	      in (ts2, ConS(k, es)) end
        | LPAR :: tr => 
              let val (ts1, e1) = expr tr
                  val ts2 = check(RPAR, ts1)
              in (ts2, e1) end
        | _ => raise Parseerror
  and defns ts = 
      let val (ts1, n) = (case ts of
                              NAME n :: tr => (tr, n)
                            | _            => raise Parseerror)
          val ts2 = check(EQUAL, ts1)
          val (ts3, e)  = expr ts2
          val (ts4, ds) = defnopt ts3
      in (ts4, (n, e) :: ds) end
  and defnopt ts =
      case ts of
          SEMI :: tr => defns tr
        | _          => (ts, [])
  and alts ts = altsopt (alt ts)
  and altsopt (ts, alt1) = 
      case ts of 
          SEMI :: tr => 
              let val (ts1, altr) = alts tr
              in (ts1, alt1 :: altr) end
        | _ => (ts, [alt1])
  and alt ts =
      case ts of
          LT :: INT k :: GT :: tr => 
              let val (ts1, vs)  = vars tr
                  val ts2 = check(DASHARROW, ts1)
                  val (ts3, rhs) = expr ts2
              in (ts3, (k, vs, rhs)) end
        | _ => raise Parseerror
  and vars ts =
      case ts of
          NAME n :: tr => 
              let val (ts1, vr) = vars tr
              in (ts1, n :: vr) end
        | _ => (ts, [])
  and exprs ts =
      case ts of 
	  COMMA :: tr => 
	      let val (ts1, e1) = expr tr
		  val (ts2, es) = exprs ts1
	      in (ts2, e1 :: es) end
	| _ => (ts, [])

  (* Scanning *)

  fun scan s =
      let exception Scanerror
          fun isblank c  = (c = #" " orelse c = #"\t" orelse c = #"\n")
          fun isdigit c  = (#"0" <= c andalso c <= #"9")
          fun islower c  = (#"a" <= c andalso c <= #"z")
          fun isupper c  = (#"A" <= c andalso c <= #"Z")
          fun isletter c = (islower c orelse isupper c)
          fun isletterdigit c = (isletter c orelse isdigit c)
          fun scname (cs, value) =
              case cs of 
                  c :: cr => if isletterdigit c then scname(cr, c :: value)
                             else (cs, implode (rev value))
                | []      => (cs, implode (rev value))
          fun intval c = ord c - ord #"0"
          fun scint (cs, value) = 
              case cs of 
                  c :: cr => if isdigit c then scint(cr, 10*value+intval c)
                             else (cs, value)
                | []      => (cs, value)
          fun skipcomment (cs, lvl) =
              case cs of
                  #"*":: #")"::cr => if lvl = 1 then cr 
				     else skipcomment(cr, lvl-1) 
                | #"(":: #"*"::cr => skipcomment(cr, lvl+1)
                | _ :: cr         => skipcomment(cr, lvl)
                | []              => raise Scanerror

          fun sc cs = 
          case cs of
              []              => []
            | #"(":: #"*"::cr => sc (skipcomment (cr, 1))
            | #"-":: #">"::cr => DASHARROW :: sc cr
            | #"=" :: cr      => EQUAL :: sc cr
            | #"+" :: cr      => ADD :: sc cr
            | #"-" :: cr      => SUB :: sc cr
            | #"*" :: cr      => MUL :: sc cr
            | #"/" :: cr      => DIV :: sc cr
            | #"%" :: cr      => MOD :: sc cr
            | #"<":: #"="::cr => LE :: sc cr     
            | #">":: #"="::cr => GE :: sc cr     
            | #"~":: #"="::cr => NE :: sc cr
            | #"<" :: cr      => LT :: sc cr     
            | #">" :: cr      => GT :: sc cr     
            | #"(" :: cr      => LPAR :: sc cr     
            | #")" :: cr      => RPAR :: sc cr     
            | #"{" :: cr      => LBRACE :: sc cr     
            | #"}" :: cr      => RBRACE :: sc cr     
            | #";" :: cr      => SEMI :: sc cr     
            | #"," :: cr      => COMMA :: sc cr     
            | #"." :: cr      => DOT :: sc cr     
            | #"|" :: cr      => BAR :: sc cr
            | #"&" :: cr      => AMPERSAND :: sc cr
            | #"\\" :: cr     => LAMBDA :: sc cr     
            | #"~" :: c :: cr => if isdigit c then
                                  let val (cs1, i) = scint(cr, intval c)
                                  in INT (~i) :: sc cs1 end
                                else raise Scanerror
            | c :: cr        => if isblank c then sc cr
                                else if isdigit c then
                                     let val (cs1, i)  = scint(cr, intval c)
                                     in INT i :: sc cs1 end
                                else if isletter c then                    
                                     let val (cs1, n) = scname(cr, [c])    
                                     in case n of
                                         "let"    => LET    :: sc cs1
                                       | "letrec" => LETREC :: sc cs1
                                       | "in"     => IN     :: sc cs1
                                       | "case"   => CASE   :: sc cs1
                                       | "of"     => OF     :: sc cs1
                                       | "if"     => IF     :: sc cs1
                                       | "then"   => THEN   :: sc cs1
                                       | "else"   => ELSE   :: sc cs1
                                       | "pack"   => PACK   :: sc cs1
                                       | "end"    => END    :: sc cs1
                                       | _        => NAME n :: sc cs1
                                     end
                                else raise Scanerror
      in sc (explode s) end
in
    fun parse cs =
	case expr (scan cs) of
	    ([], e) => e
	  | _       => raise Parseerror
		
    fun readfile (filename: string): string =
	let open TextIO
	    val is = openIn filename 
	in inputAll is before closeIn is end
end

end
